@@ -2,9 +2,11 @@
 # Copyright 2019 TD Ameritrade. Released under the terms of the 3-Clause BSD license.  # noqa: E501
 # STUMPY is a trademark of TD Ameritrade IP Company, Inc. All rights reserved.
 
+from functools import partial
 import numpy as np
 from numba import njit, prange
 from scipy.signal import convolve
+from scipy.ndimage.filters import maximum_filter1d, minimum_filter1d
 import tempfile
 import math
 
@@ -37,6 +39,20 @@ def _gpu_aamp_driver_not_found(*args, **kwargs):  # pragma: no cover
     driver_not_found()
 
 
+def _gpu_ostinato_driver_not_found(*args, **kwargs):  # pragma: no cover
+    """
+    Dummy function to raise CudaSupportError driver not found error.
+    """
+    driver_not_found()
+
+
+def _gpu_aamp_ostinato_driver_not_found(*args, **kwargs):  # pragma: no cover
+    """
+    Dummy function to raise CudaSupportError driver not found error.
+    """
+    driver_not_found()
+
+
 def _gpu_mpdist_driver_not_found(*args, **kwargs):  # pragma: no cover
     """
     Dummy function to raise CudaSupportError driver not found error.
@@ -169,7 +185,7 @@ def transpose_dataframe(df):  # pragma: no cover
     Returns
     -------
     output : df
-        If a is a Pandas `DataFrame` then return `a.T`. Otherwise, return `a`
+        If `df` is a Pandas `DataFrame` then return `df.T`. Otherwise, return `df`
     """
     if type(df).__name__ == "DataFrame":
         return df.T
@@ -442,6 +458,124 @@ def rolling_nanstd(a, w):
     )
 
 
+def _rolling_nanmin_1d(a, w=None):
+    """
+    Compute the rolling min for 1-D while ignoring NaNs.
+
+    This essentially replaces:
+
+        `np.nanmin(rolling_window(T[..., start:stop], m), axis=T.ndim)`
+
+    Parameters
+    ----------
+    a : ndarray
+        The input array
+
+    w : ndarray, default None
+        The rolling window size
+
+    Returns
+    -------
+    output : ndarray
+        Rolling window nanmin.
+    """
+    if w is None:
+        w = a.shape[0]
+
+    half_window_size = int(math.ceil((w - 1) / 2))
+    return minimum_filter1d(a, size=w)[
+        half_window_size : half_window_size + a.shape[0] - w + 1
+    ]
+
+
+def _rolling_nanmax_1d(a, w=None):
+    """
+    Compute the rolling max for 1-D while ignoring NaNs.
+
+    This essentially replaces:
+
+        `np.nanmax(rolling_window(T[..., start:stop], m), axis=T.ndim)`
+
+    Parameters
+    ----------
+    a : ndarray
+        The input array
+
+    w : ndarray, default None
+        The rolling window size
+
+    Returns
+    -------
+    output : ndarray
+        Rolling window nanmax.
+    """
+    if w is None:
+        w = a.shape[0]
+
+    half_window_size = int(math.ceil((w - 1) / 2))
+    return maximum_filter1d(a, size=w)[
+        half_window_size : half_window_size + a.shape[0] - w + 1
+    ]
+
+
+def rolling_nanmin(a, w):
+    """
+    Compute the rolling min for 1-D and 2-D arrays while ignoring NaNs.
+
+    This a convenience wrapper around `_rolling_nanmin_1d`.
+
+    This essentially replaces:
+
+        `np.nanmin(rolling_window(T[..., start:stop], m), axis=T.ndim)`
+
+    Parameters
+    ----------
+    a : ndarray
+        The input array
+
+    w : ndarray
+        The rolling window size
+
+    Returns
+    -------
+    output : ndarray
+        Rolling window nanmin.
+    """
+    axis = a.ndim - 1  # Account for rolling
+    return np.apply_along_axis(
+        lambda a_row, w: _rolling_nanmin_1d(a_row, w), axis=axis, arr=a, w=w
+    )
+
+
+def rolling_nanmax(a, w):
+    """
+    Compute the rolling max for 1-D and 2-D arrays while ignoring NaNs.
+
+    This a convenience wrapper around `_rolling_nanmax_1d`.
+
+    This essentially replaces:
+
+        `np.nanmax(rolling_window(T[..., start:stop], m), axis=T.ndim)`
+
+    Parameters
+    ----------
+    a : ndarray
+        The input array
+
+    w : ndarray
+        The rolling window size
+
+    Returns
+    -------
+    output : ndarray
+        Rolling window nanmax.
+    """
+    axis = a.ndim - 1  # Account for rolling
+    return np.apply_along_axis(
+        lambda a_row, w: _rolling_nanmax_1d(a_row, w), axis=axis, arr=a, w=w
+    )
+
+
 def compute_mean_std(T, m):
     """
     Compute the sliding mean and standard deviation for the array `T` with
@@ -861,6 +995,36 @@ def mass(Q, T, M_T=None, Σ_T=None):
     return distance_profile
 
 
+def _mass_distance_matrix(Q, T, m, distance_matrix):
+    """
+    Compute the full distance matrix between all of the subsequences of `Q` and `T`
+    using the MASS algorithm
+
+    Parameters
+    ----------
+    Q : ndarray
+        Query array
+
+    T : ndarray
+        Time series or sequence
+
+    m : int
+        Window size
+
+    distance_matrix : ndarray
+        The full output distance matrix. This is mandatory since it may be reused.
+
+    Returns
+    -------
+        None
+    """
+    k, l = distance_matrix.shape
+    T, M_T, Σ_T = preprocess(T, m)
+
+    for i in range(k):
+        distance_matrix[i, :] = mass(Q[i : i + m], T, M_T, Σ_T)
+
+
 @njit(fastmath=True)
 def _mass_absolute(Q_squared, T_squared, QT):
     """
@@ -955,6 +1119,35 @@ def mass_absolute(Q, T):
     return distance_profile
 
 
+def _mass_absolute_distance_matrix(Q, T, m, distance_matrix):
+    """
+    Compute the full non-normalized (i.e., without z-normalization) distance matrix
+    between all of the subsequences of `Q` and `T` using the MASS absolute algorithm
+
+    Parameters
+    ----------
+    Q : ndarray
+        Query array
+
+    T : ndarray
+        Time series or sequence
+
+    m : int
+        Window size
+
+    distance_matrix : ndarray
+        The full output distance matrix. This is mandatory since it may be reused.
+
+    Returns
+    -------
+    None
+    """
+    k, l = distance_matrix.shape
+
+    for i in range(k):
+        distance_matrix[i, :] = mass_absolute(Q[i : i + m], T)
+
+
 def _get_QT(start, T_A, T_B, m):
     """
     Compute the sliding dot product between the query, `T_A`, (from
@@ -1042,6 +1235,7 @@ def preprocess(T, m):
         Rolling standard deviation
     """
     T = T.copy()
+    T = transpose_dataframe(T)
     T = np.asarray(T)
     check_dtype(T)
 
@@ -1080,6 +1274,7 @@ def preprocess_non_normalized(T, m):
         `np.nan`/`np.inf` value (False)
     """
     T = T.copy()
+    T = transpose_dataframe(T)
     T = np.asarray(T)
     check_dtype(T)
 
@@ -1306,3 +1501,42 @@ def rolling_isfinite(a, w):
     a_subseq_isfinite[~a_isfinite[w - 1 :]] = False
 
     return a_isfinite[: a_isfinite.shape[0] - w + 1]
+
+
+def _get_partial_mp_func(mp_func, dask_client=None, device_id=None):
+    """
+    A convenience function for creating a `functools.partial` matrix profile function
+    for single server (parallel CPU), multi-server with Dask distributed (parallel CPU),
+    and multi-GPU implementations.
+
+    Parameters
+    ----------
+    mp_func : object
+        The matrix profile function to be used for computing a matrix profile
+
+    dask_client : client, default None
+        A Dask Distributed client that is connected to a Dask scheduler and
+        Dask workers. Setting up a Dask distributed cluster is beyond the
+        scope of this library. Please refer to the Dask Distributed
+        documentation.
+
+    device_id : int or list, default None
+        The (GPU) device number to use. The default value is `0`. A list of
+        valid device ids (int) may also be provided for parallel GPU-STUMP
+        computation. A list of all valid device ids can be obtained by
+        executing `[device.id for device in numba.cuda.list_devices()]`.
+
+    Returns
+    -------
+    partial_mp_func : object
+        A generic matrix profile function that wraps the `dask_client` or GPU
+        `device_id` into `functools.partial` function where possible
+    """
+    if dask_client is not None:
+        partial_mp_func = partial(mp_func, dask_client)
+    elif device_id is not None:
+        partial_mp_func = partial(mp_func, device_id=device_id)
+    else:
+        partial_mp_func = mp_func
+
+    return partial_mp_func
