@@ -5,13 +5,104 @@
 import logging
 
 import numpy as np
+from scipy.stats import norm
 from numba import njit, prange
+from functools import lru_cache
 
 from . import core
 
 logger = logging.getLogger(__name__)
 
 
+def _preprocess_include(include):
+    """
+    A utility function for processing the `include` input
+
+    Parameters
+    ----------
+    include : ndarray
+        A list of (zero-based) indices corresponding to the dimensions in `T` that
+        must be included in the constrained multidimensional motif search.
+        For more information, see Section IV D in:
+
+        `DOI: 10.1109/ICDM.2017.66 \
+        <https://www.cs.ucr.edu/~eamonn/Motif_Discovery_ICDM.pdf>`__
+
+    Returns
+    -------
+    include : ndarray
+        Process `include` and remove any redundant index values
+    """
+    include = np.asarray(include)
+    _, idx = np.unique(include, return_index=True)
+    if include.shape[0] != idx.shape[0]:  # pragma: no cover
+        logger.warning("Removed repeating indices in `include`")
+        include = include[np.sort(idx)]
+
+    return include
+
+
+def _apply_include(
+    D,
+    include,
+    restricted_indices=None,
+    unrestricted_indices=None,
+    mask=None,
+    tmp_swap=None,
+):
+    """
+    Apply a transformation to the multi-dimensional distance profile so that specific
+    dimensions are always included. Essentially, it is swapping rows within the distance
+    profile.
+
+    Parameters
+    ----------
+    D : ndarray
+        The multi-dimensional distance profile
+
+    include : ndarray
+        A list of (zero-based) indices corresponding to the dimensions in `T` that
+        must be included in the constrained multidimensional motif search.
+        For more information, see Section IV D in:
+
+        `DOI: 10.1109/ICDM.2017.66 \
+        <https://www.cs.ucr.edu/~eamonn/Motif_Discovery_ICDM.pdf>`__
+
+    restricted_indices : ndarray, default None
+        A list of indices specified in `include` that reside in the first
+        `include.shape[0]` rows
+
+    unrestricted_indices : ndarray, default None
+        A list of indices specified in `include` that do not reside in the first
+        `include.shape[0]` rows
+
+    mask : ndarray, default None
+        A boolean mask to select for unrestricted indices
+
+    tmp_swap : ndarray, default None
+        A reusable array to aid in array element swapping
+    """
+    include = _preprocess_include(include)
+
+    if restricted_indices is None:
+        restricted_indices = include[include < include.shape[0]]
+
+    if unrestricted_indices is None:
+        unrestricted_indices = include[include >= include.shape[0]]
+
+    if mask is None:
+        mask = np.ones(include.shape[0], bool)
+        mask[restricted_indices] = False
+
+    if tmp_swap is None:
+        tmp_swap = D[: include.shape[0]].copy()
+    else:
+        tmp_swap[:] = D[: include.shape[0]]
+
+    D[: include.shape[0]] = D[include]
+    D[unrestricted_indices] = tmp_swap[mask]
+
+
 def _multi_mass(Q, T, m, M_T, Σ_T, μ_Q, σ_Q):
     """
     A multi-dimensional wrapper around "Mueen's Algorithm for Similarity Search"
@@ -59,25 +150,73 @@ def _multi_mass(Q, T, m, M_T, Σ_T, μ_Q, σ_Q):
     return D
 
 
-def _apply_include(
-    D,
-    include,
-    restricted_indices=None,
-    unrestricted_indices=None,
-    mask=None,
-    tmp_swap=None,
-):
+@lru_cache()
+def _inverse_norm(n_bit=8):
     """
-    Apply a transformation to the multi-dimensional distance profile so that specific
-    dimensions are always included. Essentially, it is swapping rows within the distance
-    profile.
+    Generate bin edges from an inverse normal distribution
 
     Parameters
     ----------
-    D : ndarray
-        The multi-dimensional distance profile
+    n_bit : int, default 8
+        The number of bits to be used in generating the inverse normal distribution
 
-    include : ndarray
+    Returns
+    -------
+    out : ndarray
+        Array of bin edges that can be used for data discretization
+    """
+    return norm.ppf(np.arange(1, (2 ** n_bit)) / (2 ** n_bit))
+
+
+def _discretize(a, bins, right=True):
+    """
+    Discretize each row of the input array
+
+    Parameters
+    ----------
+    a : ndarray
+        The input array
+
+    bins : ndarray
+        The bin edges used to discretize `a`
+
+    right : bool, default True
+        Indicates whether the intervals for binning include the right or the left bin
+        edge.
+
+    Returns
+    -------
+    out : ndarray
+        Discretized array
+    """
+    return np.digitize(a, bins, right=right)
+
+
+def subspace(T, m, subseq_idx, nn_idx, k, include=None, discords=False):
+    """
+    Compute the k-dimensional matrix profile subspace for a given subsequence index and
+    its nearest neighbor index
+
+    Parameters
+    ----------
+    T : ndarray
+        The time series or sequence for which the multi-dimensional matrix profile,
+        multi-dimensional matrix profile indices were computed
+
+    m : int
+        Window size
+
+    subseq_idx : int
+        The subsequence index in T
+
+    nn_idx : int
+        The nearest neighbor index in T
+
+    k : int
+        The subset number of dimensions out of `D = T.shape[0]`-dimensions to return
+        the subspace for
+
+    include : ndarray, default None
         A list of (zero-based) indices corresponding to the dimensions in `T` that
         must be included in the constrained multidimensional motif search.
         For more information, see Section IV D in:
@@ -85,60 +224,69 @@ def _apply_include(
         `DOI: 10.1109/ICDM.2017.66 \
         <https://www.cs.ucr.edu/~eamonn/Motif_Discovery_ICDM.pdf>`__
 
-    restricted_indices : ndarray, default None
-        A list of indices specified in `include` that reside in the first
-        `include.shape[0]` rows
+    discords : bool, default False
+        When set to `True`, this reverses the distance profile to favor discords rather
+        than motifs. Note that indices in `include` are still maintained and respected.
 
-    unrestricted_indices : ndarray, default None
-        A list of indices specified in `include` that do not reside in the first
-        `include.shape[0]` rows
-
-    mask : ndarray, default None
-        A boolean mask to select for unrestricted indices
-
-    tmp_swap : ndarray, default None
-        A reusable array to aid in array element swapping
+    Returns
+    -------
+        S : ndarray
+        An array of that contains the `k`th-dimensional subspace for the subsequence
+        with index equal to `motif_idx`
     """
-    if restricted_indices is None:
-        restricted_indices = include[include < include.shape[0]]
+    T, _, _ = core.preprocess(T, m)
 
-    if unrestricted_indices is None:
-        unrestricted_indices = include[include >= include.shape[0]]
+    subseqs = core.z_norm(T[:, subseq_idx : subseq_idx + m], axis=1)
+    neighbors = core.z_norm(T[:, nn_idx : nn_idx + m], axis=1)
+    D = np.linalg.norm(subseqs - neighbors, axis=1)
 
-    if mask is None:
-        mask = np.ones(include.shape[0], bool)
-        mask[restricted_indices] = False
-
-    if tmp_swap is None:
-        tmp_swap = D[: include.shape[0]].copy()
+    if discords:
+        sorted_idx = D[::-1].argsort(axis=0, kind="mergesort")
     else:
-        tmp_swap[:] = D[: include.shape[0]]
+        sorted_idx = D.argsort(axis=0, kind="mergesort")
 
-    D[: include.shape[0]] = D[include]
-    D[unrestricted_indices] = tmp_swap[mask]
+    # `include` processing occur here since we are dealing with indices, not distances
+    if include is not None:
+        include = _preprocess_include(include)
+        mask = np.in1d(sorted_idx, include)
+        include_idx = mask.nonzero()[0]
+        exclude_idx = (~mask).nonzero()[0]
+        sorted_idx[: include_idx.shape[0]], sorted_idx[include_idx.shape[0] :] = (
+            sorted_idx[include_idx],
+            sorted_idx[exclude_idx],
+        )
+
+    S = sorted_idx[: k + 1]
+
+    n_bit = 8
+    bins = _inverse_norm()
+    disc_subseqs = _discretize(subseqs[S], bins)
+    disc_neighbors = _discretize(neighbors[S], bins)
+    n_val = np.unique(disc_subseqs - disc_neighbors).shape[0]
+    bit_size = n_bit * (T.shape[0] * m * 2 - k * m)
+    bit_size = bit_size + k * m * np.log2(n_val) + n_val * n_bit
+
+    return S
 
 
 def _query_mstump_profile(
     query_idx, T_A, T_B, m, excl_zone, M_T, Σ_T, μ_Q, σ_Q, include=None, discords=False
 ):
     """
-    Multi-dimensional wrapper to compute the multi-dimensional matrix profile,
-    the multi-dimensional matrix profile index, the multi-dimensional matrix profile
-    subspace for a given query window within the times series or sequence that is
-    denoted by the `query_idx` index. Essentially, this is a convenience wrapper around
-    `_multi_mass`.
+    Multi-dimensional wrapper to compute the multi-dimensional matrix profile and
+    the multi-dimensional matrix profile index for a given query window within the times
+    series or sequence that is denoted by the `query_idx` index. Essentially, this is a
+    convenience wrapper around `_multi_mass`.
 
     Parameters
     ----------
     query_idx : int
-        The window index to calculate the first multi-dimensional matrix profile,
-        multi-dimensional matrix profile indices, and multi-dimensional matrix profile
-        subspace.
+        The window index to calculate the first multi-dimensional matrix profile and
+        multi-dimensional matrix profile indices
 
     T_A : ndarray
-        The time series or sequence for which the multi-dimensional matrix profile,
-        multi-dimensional matrix profile indices, and multi-dimensional matrix profile
-        subspace will be returned
+        The time series or sequence for which the multi-dimensional matrix profile and
+        multi-dimensional matrix profile indices
 
     T_B : ndarray
         The time series or sequence that contains your query subsequences
@@ -182,13 +330,6 @@ def _query_mstump_profile(
     I : ndarray
         Multi-dimensional matrix profile indices for the window with index
         equal to `query_idx`
-
-    S : ndarray
-        A ragged array of ndarrays that contain the multi-dimensional subspace for the
-        window with index equal to `query_idx`. The `len(S)` will be equal to the total
-        number of dimensions, `d`, and where `S[i]` corresponds to the list of subspace
-        indices for the `i+1`th subspace dimension (i.e., `S[1]` corresponds to the
-        subspace with dimension `2` and with `len(S[1]) == 2`).
     """
     d, n = T_A.shape
     k = n - m + 1
@@ -208,15 +349,9 @@ def _query_mstump_profile(
         start_row_idx = include.shape[0]
 
     if discords:
-        # D[start_row_idx:][::-1].sort(axis=0)
-        sorted_idx = D[start_row_idx:][::-1].argsort(axis=0, kind="mergesort")
-        broadcast_idx = np.arange(D[start_row_idx:].shape[1])[np.newaxis, :]
-        D[start_row_idx:][::-1] = D[start_row_idx:][::-1][sorted_idx, broadcast_idx]
+        D[start_row_idx:][::-1].sort(axis=0, kind="mergesort")
     else:
-        # D[start_row_idx:].sort(axis=0)
-        sorted_idx = D[start_row_idx:].argsort(axis=0, kind="mergesort")
-        broadcast_idx = np.arange(D[start_row_idx:].shape[1])[np.newaxis, :]
-        D[start_row_idx:] = D[start_row_idx:][sorted_idx, broadcast_idx]
+        D[start_row_idx:].sort(axis=0, kind="mergesort")
 
     D_prime = np.zeros(k)
     for i in range(d):
@@ -227,18 +362,15 @@ def _query_mstump_profile(
 
     P = np.full(d, np.inf, dtype="float64")
     I = np.full(d, -1, dtype="int64")
-    S = np.empty(d, dtype=object)
 
     for i in range(d):
         min_index = np.argmin(D[i])
         I[i] = min_index
         P[i] = D[i, min_index]
-        S[i] = sorted_idx[: i + 1, min_index]
         if np.isinf(P[i]):  # pragma nocover
             I[i] = -1
-            S[i][:] = -1
 
-    return P, I, S
+    return P, I
 
 
 def _get_first_mstump_profile(
@@ -306,7 +438,7 @@ def _get_first_mstump_profile(
         Multi-dimensional matrix profile indices for the window with index
         equal to `start`
     """
-    P, I, _ = _query_mstump_profile(
+    P, I = _query_mstump_profile(
         start, T_A, T_B, m, excl_zone, M_T, Σ_T, μ_Q, σ_Q, include, discords
     )
     return P, I
@@ -620,6 +752,7 @@ def _mstump(
             σ_Q,
         )
 
+        # `include` processing must occur here since we are dealing with distances
         if include is not None:
             _apply_include(
                 D,
@@ -678,12 +811,12 @@ def mstump(T, m, include=None, discords=False):
     Returns
     -------
     P : ndarray
-        The multi-dimensional matrix profile. Each column of the array corresponds
-        to each matrix profile for a given dimension (i.e., the first column is
-        the 1-D matrix profile and the second column is the 2-D matrix profile).
+        The multi-dimensional matrix profile. Each row of the array corresponds
+        to each matrix profile for a given dimension (i.e., the first row is
+        the 1-D matrix profile and the second row is the 2-D matrix profile).
 
     I : ndarray
-        The multi-dimensional matrix profile index where each column of the array
+        The multi-dimensional matrix profile index where each row of the array
         corresponds to each matrix profile index for a given dimension.
 
     Notes
@@ -693,7 +826,7 @@ def mstump(T, m, include=None, discords=False):
 
     See mSTAMP Algorithm
     """
-    T_A = core.transpose_dataframe(T)
+    T_A = T
     T_B = T_A
 
     T_A, M_T, Σ_T = core.preprocess(T_A, m)
@@ -706,11 +839,7 @@ def mstump(T, m, include=None, discords=False):
     core.check_window_size(m)
 
     if include is not None:
-        include = np.asarray(include)
-        _, idx = np.unique(include, return_index=True)
-        if include.shape[0] != idx.shape[0]:  # pragma: no cover
-            logger.warning("Removed repeating indices in `include`")
-            include = include[np.sort(idx)]
+        include = _preprocess_include(include)
 
     d, n = T_B.shape
     k = n - m + 1
@@ -745,4 +874,4 @@ def mstump(T, m, include=None, discords=False):
         discords,
     )
 
-    return P.T, I.T
+    return P, I
