@@ -6,7 +6,8 @@ import logging
 
 import numpy as np
 
-from . import core, _stump
+from . import core
+from .stump import _stump
 
 logger = logging.getLogger(__name__)
 
